############################################################
#                                                          #
#                   load libraries                         #
#                                                          #
############################################################
library(sp)
library(ggmap)
library(raster)
library(mapview)
library(viridis)
library(akima)

crs_model        <- "+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0" #coordinates sistem

pollutants       <- c("pm10","pm2p5","no2","o3")
years            <- c(2020,2021,2022) 

general_path     <- "/esarchive/scratch/mhajji/Meriem/Gene/results_regional_"
save_png         <- "/esarchive/scratch/mhajji/Meriem/TFM_matrix/maps_TFM/"


############################################################
#                                                          #
#             load funtion   ggplot                        #
#                                                          #
############################################################

plot_function <- function(raster, zlim2, plotname, breaks, legend_texting, labels){
  raster[raster$"var1.pred" > zlim2[2]]  <- zlim2[2]
  raster[raster$"var1.pred" < zlim2[1]]  <- zlim2[1]
  
  plot <- ggmap(map_transparent1) +
    coord_sf()+
    geom_raster(data = as.data.frame(raster, xy = TRUE) , aes(x = x, y = y,fill = "var1.pred"), alpha = 0.85)+ 
    theme_linedraw()  +
    scale_fill_gradientn(colors = cr, breaks = breaks, guide = guide_colorbar(frame.colour = "black", ticks.colour = "black"),
                         limits = zlim2, labels = labels ) +
    theme(
      panel.grid.major = element_blank(),
      panel.grid.minor = element_blank(),
      text = element_text(size = 20),
      axis.title.x = element_text("lon"), axis.title.y = element_text("lat")) + 
    labs(fill = legend_texting) 
  png(file = paste0(save_png,plotname, ".png"), width = 1200, height = 1200)
  print(plot)
  dev.off()
}


############################################################
#                                                          #
#              create background map                       #
#                                                          #
############################################################

#To save the ggmap object, the following lines have to be run with internet connection
sbbox                        <- make_bbox(lon = c(-0.001190186,3.497162), lat = c(40.24870,43.02355), f = 0)  # TODO: Integrar amb el configfile
map1                         <- get_stamenmap(sbbox, maptype = "terrain-background", zoom = 9)
bw                           <- ggmap(map1)
mapatt1                      <- attributes(map1)
map_transparent1             <- matrix(adjustcolor(map1, alpha.f = 0.6))
attributes(map_transparent1) <- mapatt1

############################################################
#                                                          #
#                   load file to plot                      #
#                                                          #
############################################################


for(p in 1:length(pollutants)){
  pollutant <- pollutants[p]
  
  if(pollutant=="no2"){
    rang    <- c(0,10,20,30,40,80,120,160,200,300,400,410)
    zlim2   <- c(0,410)
    breaks  <- rang
  } else if(pollutant=="o3"){
    rang    <- c(0,60,80,110,140,160,170,180,190,240,250)
    zlim2   <- c(0,250)
    breaks  <- rang
  } else if(pollutant=="pm10" | pollutant=="pm2p5"){
    rang    <- c(0,10,20,30,35,40,45,50,60,70,80,90)
    zlim2   <- c(0,90)
    breaks  <- rang 
  }
  
  labels         <- breaks
  
  for(y in 1:length(years)){
    year            <- years[y]
    
    if(pollutant %in% c("no2","pm10","pm2p5")){
      annual_mean_name <- paste0("df_annual_mean_",year,"_",pollutant,"_df.rds")
      annual_max_name  <- paste0("df_annual_max_",year,"_",pollutant,"_df.rds")
    }else if(pollutant=="o3"){
      annual_mean_name <- paste0("df_annual_mean_",year,"_",pollutant,"_df.rds")
      annual_max_name  <- paste0("df_oct_max_",year,"_",pollutant,"_df.rds")
    }
    
    pollutant_path  <- paste0(general_path,year,"/results_gene_",pollutant)
    
    sp_annual_mean <- readRDS(file=paste0(pollutant_path,"/",annual_mean_name))
    sp_annual_max  <- readRDS(file=paste0(pollutant_path,"/",annual_max_name))
    
    sp_annual_mean <- spTransform(sp_annual_mean, sp::CRS(crs_model))
    sp_annual_max  <- spTransform(sp_annual_max, sp::CRS(crs_model))
    
    ############################################################
    #                                                          #
    #   make raster in a uniform grid for annual mean          #
    #                                                          #
    ############################################################
    mean_uniform           <- interp(x=sp_annual_mean,z="pollutant",nx=300,ny=300)
    mean_raster            <- raster::raster(mean_uniform$z)
    mean_raster@data@names <- "var1.pred"
    raster                 <- mean_raster

    
    
    cr=viridis(256) #aixo era magma(200)
    # library(viridis)  # Asegúrate de tener instalado el paquete 'viridis'
    # 
    # cr <- viridis(256, option = "plasma")
    
    # zlim2          <- c(15,25)
    # breaks         <- seq(zlim2[1], zlim2[2],2) #c(15,20,21,25)
    # labels         <- breaks
    legend_texting <- expression(paste(pollutant, '[' ,mu, 'g/', m^3 ,']')) #"PM10 ug/m3"
    plotname       <- paste0(pollutant,"_annual_mean")
    
    plot_function(raster, zlim2, plotname, breaks, legend_texting, labels)
    
    
    ############################################################
    #                                                          #
    #   make raster in a uniform grid for annual "i" max       #
    #                                                          #
    ############################################################

   
    mean_uniform_max         <- interp(x=sp_annual_max,z="pollutant",nx=300,ny=300)
    mean_raster_max          <- raster::raster(mean_uniform_max[,'pollutant'])
    mean_raster_max@data@names   <- ""var1.pred""
    raster_max                   <- mean_raster_max
    
    cr=magma(200)
    
    # zlim2_max          <- c(15,25)
    # breaks_max         <- seq(zlim2[1], zlim2[2],2) 
    # labels_max         <- breaks
    
    legend_texting_max <- expression(paste(pollutant, '[' ,mu, 'g/', m^3 ,']')) #"PM10 ug/m3"
    plotname_max       <- paste0(pollutant,"_annual_mean")
    
    plot_function(raster_max, zlim2, plotname_max, breaks, legend_texting_max, labels)
    
    }
}




