# Needs packages:
#library(urbanR)
library(gbm)
library(lubridate)
library(plyr) #OJO! la llibreria plyr ha de ser cridada abans que dplyr 
library(dplyr) #per treballar amb cbinds i esas cosas
library(caret)
library(zoo)
#library(scatterplot3d)
#library(plotly)

#borrar
#library(RNetCDF)
library(sp)
library(sf)
#library(mapview)
#library(ncdf4)
#library(rgeos) 
library(ggplot2)
#library(broom)
#library(magrittr)
#library(ggpubr)
#library(raster)
#library(rgdal)
#library(tmap)
library(gstat)
#library(abind)
#library(rbind)
library(openair)
#library(tidyr)



############################################################
#                                                          #
#                    Initial data                          #
#                                                          #
############################################################
pollutant       <- "pm2p5"   #no2, o3, pm10
years           <- c(2019, 2021, 2022)
#years           <- 2019
#save_result     <-"/esarchive/scratch/abarrant/results_regional_"   

#ASSEGURAR QUE NO ES GUARDA RES A SAVE_RESULT
extract_result     <-"/esarchive/scratch/ccarnere/RScripts/results_regional_"
file_format     <-"gene"   #ghost
time_resolution <-"daily"  #"hourly"
compute_in_UTM  <- "TRUE"
crs_latlon      <- "+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0"
crs_utm         <- "+proj=utm +zone=31 +ellps=intl +units=m +no_defs"

save_path    <- "/esarchive/scratch/abarrant/results_regional_2019/GBM_eval/"  ##change

############################################################
#                                                          #
#                 Load Local Stations                      #
#                                                          #
############################################################
#We will pick the latest data available
local_stations     <- readRDS(paste0(extract_result,years[length(years)],"/results_gene_",pollutant, "/output/","local_stations_",file_format, "_",pollutant,".rds"))

############################################################
#                                                          #
#                  Load All observations                   #
#                                                          #
############################################################

df_obs_years <- NULL
for(read_year in years){
  df_obs_all        <- readRDS(file = paste0(extract_result,read_year,"/results_gene_",pollutant, "/output/","df_obs_",file_format,"_",pollutant,".rds"))
  
  start_date  <- as.Date(paste(read_year, "-01-01", sep = ""))
  end_date    <- as.Date(paste(read_year, "-12-31", sep = ""))
  
  dates      <- seq(start_date, end_date, by = "day")                   
  df_obs_sub <- subset(df_obs_all, date>=start_date & date <= end_date) 
  
  
  # Comprovation we don't have duplications in obs
  if(any(duplicated(df_obs_sub$date))){
    print(paste0("Warning, there is duplicated observations on date :", df_obs_sub$date[duplicated(df_obs_sub$date)]))
    df_obs_sub <- distinct(df_obs_sub, date, .keep_all = TRUE)
    print("The df_obs_sub (alias=observations) have been modified to have the same length as the model")
  }
  
  if (length(df_obs_sub$date) != length(dates)){
    print("WARNING: There are missing hours/dates in observations. And the code is not developed for daily data yet!!")
  }
  
  df_obs_years <- rbind.fill(df_obs_years, df_obs_sub) #combine df by row, filling in missing columns
}

names_local_stations <- as.list(intersect(local_stations@data$ref_obs, colnames(df_obs_years)))

############################################################
#                                                          #
#            Load raw model interpolated to AQMS           #
#                                                          #
############################################################
df_all_model_raw_at_AQMS <- NULL

for(read_year in years){
  model_raw_at_AQMS_df     <- readRDS(paste0(extract_result,read_year,"/results_gene_",pollutant, "/output/","model_raw_at_AQMS_",file_format, "_",pollutant,"_df.rds"))
  model_raw_at_AQMS_df     <- mutate(model_raw_at_AQMS_df, date = as.Date(date))
  df_all_model_raw_at_AQMS <- rbind.fill(df_all_model_raw_at_AQMS, model_raw_at_AQMS_df)
}





obs_filled       <- df_obs_years 
AQMS_obs         <- subset(df_obs_years, select = -date) #we remove the column dates from observations data frame
OK_other_obs_df  <- NULL
list_gbm_results <- NULL
Stats_gbm <- NULL

gap_sizes <- 1:7


start_time <- Sys.time()

# To fill the gaps of observations in each station, so we need a loop to repeat the fill process in each station:
for (s in seq(13, 21, by=8)){
  #s <- 13 #eixample
  s <- 21 #la senia by 8
  station                 <- names_local_stations[[s]]
  
  # Arrangement of AQMS interpolated to AQMS
  # station is the code station that we hidded to interpolate to it, first
  col_station           <- which(station==colnames(df_obs_years))   # index de l'estació en les obs i treure-la
  stations_to_interp    <- df_obs_years[,-col_station]              # obs de les altres estacions sense l'estació a interpolar
  c <- 0 #comptador de dies que no es pot fer la interpolacio
  
  #Loop over all days to interpolate data from other station to station
  for(j in 1:length(df_obs_years$date)){
    #j <- 1
    #print(paste0("Date: ", stations_to_interp$date[j]))
    #We take the data day by day /hour by hour
    ok_obs    <- stations_to_interp[j,] #obs per dia de totes les estacions menys la eliminada
    ok_obs    <- subset(ok_obs,select=-date)
    df_ok_obs <- data.frame(ref_obs = colnames(ok_obs), pollutant = as.numeric(ok_obs)) #data frame estacions x 1 dia
    
    #coordinates where to interpolate:
    df_hidded_AQMS <- local_stations[ which(local_stations$ref_obs == station) , ] #spatial PDF de l'estació a interpolar
    
    if (compute_in_UTM) {
      df_hidded_AQMS <- sp::spTransform(df_hidded_AQMS, sp::CRS(crs_utm))
    }
    
    #    1.2 Compute Variogram
    #with other AQMS we want a SPDF that has the ref of the station and the coordinates of each, with the value of the pollution in that day/hour
    sp_other_ok_obs <- local_stations[ -which(local_stations$ref_obs == station) , ] #spatial PDF on hi han totes les coords excepte l'amagada (sobren estacions)
    df_other_obs <- as.data.frame(sp_other_ok_obs) #pasem de spdf a df per poder filtrar
    df_obs_data  <- merge(df_other_obs[, -which(names(df_other_obs) == "pollutant")], df_ok_obs, by = "ref_obs") #copiem les coordenades de les estacions d'interès amb les observacion que pertoquen
    
    if(!sum(is.na(df_obs_data$pollutant))>=length(df_obs_data$pollutant)){
      df_obs_data                   <- na.omit(df_obs_data)
      coordinates(df_obs_data)      <- ~lon+lat
      sp::proj4string(df_obs_data)  <- crs_latlon
      
      if (compute_in_UTM) {
        df_obs_data <- spTransform(df_obs_data, sp::CRS(crs_utm))
      }
      
      #efectuem l'ordinary kriging
      pollutant.fit_ok          <- automap::autofitVariogram(formula=pollutant~1, input_data=df_obs_data,model="Ste")  #stein again funciona millor per pm2p5
      
      ok_gstat <- gstat(formula=pollutant~1, data = df_obs_data, model = pollutant.fit_ok$var_model, nmax=100)
      OK       <- predict(object=ok_gstat, newdata=df_hidded_AQMS)
      
      
      OK_other_obs           <- data.frame(data=OK$var1.pred)
      colnames(OK_other_obs) <- station
      OK_other_obs$date      <- stations_to_interp$date[j]
      
      OK_other_obs_df <-  plyr::rbind.fill(as.data.frame(OK_other_obs_df), OK_other_obs)
      
    } else{
      #print(paste0("There is no observations in: ", stations_to_interp$date[j]))
      OK_other_obs           <- data.frame(data=NA)
      colnames(OK_other_obs) <- station
      OK_other_obs$date      <- stations_to_interp$date[j]
      
      OK_other_obs_df        <-  plyr::rbind.fill(as.data.frame(OK_other_obs_df), OK_other_obs)
      c <- c+1
    }
  }
  print(paste0("The number of days without any observational data is: ", c))  #dies pels quals no hi ha cap obs a cap de les estacions i doncs no pot interpolar sense dades
  
  # Select station data (PREDICTORS)
  single_station_data <- data.frame(date            = OK_other_obs_df$date,
                                    obs             = AQMS_obs[,station],                     #valor obs a l'estacio (de vegades nan)
                                    model           = df_all_model_raw_at_AQMS[,station],     #valor donat pel model a l'estacio
                                    day_of_year     = yday(OK_other_obs_df$date),             #julian day
                                    day_of_week     = wday(OK_other_obs_df$date),
                                    other_obs       = OK_other_obs_df[,station],              #predictor del valor interpolat a lestacio a partir de totes les altres
                                    year            = year(OK_other_obs_df$date))             #dia de lany
  #estacions interpolat a l'estació
  
  single_station_data                     <- na.omit(single_station_data)  #ometre predictors nan
  single_station_data <- mutate(single_station_data, date = as.Date(date))
  
  #agafo nomes l'any on fare els forats
  single_station_data_year <- single_station_data %>% 
    filter(year(date) == 2019)
  
  #genero un df sense l'any per poder posar despres les noves dades de training
  single_station_data_other_years <- single_station_data %>% 
    filter(year(date) != 2019)
  
  
  for (i_seed in 1:100){
    #i_seed <- 5
    set.seed(i_seed)
    print(i_seed)
    
    #bucle per cada gap
    for (i in seq_along(gap_sizes)) {
      
      gap_size <- gap_sizes[i]             #agafar mida gap
      #gap_start <- gap_start_values[i]     #agafar limit inferior train fraction 
      
      #gap_size <- 1
      #gap_start <- 0.7
      
      #bucle de cada train fraction 
      for (train_fraction in seq(0.1, 0.9, by = 0.1)) {
        
        #train_fraction <- 0.7
        #sample size, numero de files segons la train fraction
        
        smp_size <- floor(train_fraction * nrow(single_station_data_year))
        
        
        ###GENERAR UN SUBSET DE NOMES EL 2019, PER AGAFAR LA DATA TRAINING I TEST DEL 2019, 
        #PERO DESPRES DONARLI UN MERGE DE LA DATA TRAINING DEL SUBSET MES LA RESTA DANYS AL GBM
        
        #generem dataframes per posar les dades de les observacions i els gaps
        Obs_Train <- single_station_data_year                                        
        Obs_Test <- single_station_data_year
        Obs_Test$obs <- NA
        data_training <- single_station_data_year
        
        array_ind <- NULL
        
        #bucle per generar els gaps (fins que el nombre de files a data_training es menor a la smp_size)
        while (nrow(data_training) >= smp_size && length(unique(array_ind)) < nrow(Obs_Train)  ) {
          
          ind <- sample(seq_len(nrow(Obs_Train)), size = 1)  #agafem un index random de mida 1, del df data_training_bags, on hem exclos els valors del contron del gap
          array_ind <- c(array_ind, ind)
          #ind_date <- sample(Obs_Train$date, size = 1)
          #ind <- which(Obs_Train$date==ind_date)
          
          #condicions de bon index escollit per crear gap (ni al final, i mirant valors anteriors i posteriors interiors al gap)
          if (ind > 1 && ind <= (nrow(Obs_Train) - gap_size) && !is.na(Obs_Train$obs[ind - 1]) && !is.na(Obs_Train$obs[ind]) && !is.na(Obs_Train$obs[ind + gap_size])) {
            data_test <- Obs_Train[ind:(ind + gap_size - 1), ] %>% arrange(date)                 #agafem les dades del gap, en funcio mida gap
            Obs_Test[Obs_Test$date %in% data_test$date, "obs"] <- data_test[, "obs"]  #col.loquem les dades dels gaps
            Obs_Train[Obs_Train$date %in% data_test$date, "obs"] <- NA                     #fem el forat
            data_training <- na.omit(Obs_Train)
          }
        }
        
        #si ha arribat al limit de train_fraction, provar una altra
        if (length(unique(array_ind)) == nrow(Obs_Train) ){
          next
        }
        
        #juntem totes les dades del forat
        data_test <- na.omit(Obs_Test)
        
        data_training_all <- rbind(data_training, single_station_data_other_years)
        
        
        
        # Preliminary fit
        n.trees = 5000 #realment s'ha de provar i escollir l'òptim (si dona temps bucle amb 3-4 opcions per comparar) [la que dongui menor mse]
        Vx_formula <- "obs ~ model + day_of_year + day_of_week + other_obs + year"
        gbm1 <- gbm(as.formula(Vx_formula),
                    data = data_training_all,
                    distribution = "gaussian",
                    n.trees = n.trees,
                    interaction.depth = 1, #amb 1 hauria de funcionar, no solem posar més de 10
                    shrinkage = 0.01,      #learning rate ni molt petit pk anirà lent ni molt gran pk te pasas [ns si bucle amb 3 proves]
                    cv.folds = 5,          #cross validation
                    n.cores = NULL,        # will use all cores by default
                    verbose = FALSE)  
        gbm.fit <- gbm1
        
        # print results
        print(gbm.fit)
        
        ####comentar aixo si es fa a nord3
        # plot loss function as a result of n trees added to the ensemble (no a nord3)
        #gbm.perf(gbm.fit, method = "cv")
        
        # get RMSE 
        #sqrt(min(gbm.fit$cv.error))  #agafa el rmse minim (pero no esta a cap variable?)
        
        # Variable importance
        #gbm.fit.final <- gbm.fit
        #par(mar = c(5, 8, 1, 1))     #unable to display on nord3
        #summary(
        #gbm.fit.final, 
        #cBars = 10,
        #method = relative.influence, # also can use permutation.test.gbm
        #las = 2)
        ####
        
        # Data frame amb dos variables date(fecha + hora) i Tests_results/Obs_Test/Obs_Train (NA) inicialitzacio df amb nan
        Test_results_station  <-  data.frame(date=df_obs_years$date , Test_results = NA) 
        Obs_Test_station      <-  data.frame(date=df_obs_years$date , Obs_Test = NA)
        Obs_Train_station     <-  data.frame(date=df_obs_years$date , Obs_Train = NA)
        
        #Once you have decided on a final model you will likely want to use the model to predict on new observations. 
        res_training <- predict( gbm1 , newdata=data_training)
        res_test     <- predict( gbm1 , newdata=data_test[,c("model","day_of_year","day_of_week","other_obs", "year")])
        
        Test_results <- data.frame(date = data_test[,'date'], Test_results=res_test )
        Test_results_station[Test_results_station$date%in%Test_results$date ,"Test_results"] <- Test_results[,"Test_results"]
        
        Obs_Test <- data.frame(date = data_test[,'date'], Obs_Test=data_test$obs )
        Obs_Test_station[Obs_Test_station$date%in%Obs_Test$date ,"Obs_Test"] <- Obs_Test[,"Obs_Test"]
        
        Obs_Train <- data.frame(date = data_training[,'date'], Obs_Train=data_training$obs )
        Obs_Train_station[Obs_Train_station$date%in%Obs_Train$date ,"Obs_Train"] <- Obs_Train[,"Obs_Train"]
        
        #generem un df conjunt
        data_single     <- data.frame(date                    = df_obs_years$date,
                                      Obs_Train_station       = Obs_Train_station$Obs_Train,
                                      Obs_Test_station        = Obs_Test_station$Obs_Test,
                                      Test_results_station    = Test_results_station$Test_results,
                                      model                   = df_all_model_raw_at_AQMS[,station])
        list_gbm_results[[ s ]] <- data_single
        
        # TIME_PLOT
        #png(file = paste0(save_time_series,local_stations$ref_obs[l],"_", local_stations$station_name[l],"_",mm,".png"), width = 800, height = 400)
        #openair::timePlot(data_single,
        #         pollutant = c("Obs_Train_station","Obs_Test_station","Test_results_station","model"),
        #         group = TRUE,
        #         title=station,
        #         plot.type= "h",
        #         y.relation = "free",
        #         ylab="PM2,5 (ug/m3)",
        #         cols = c("black", "blue", "red", "green"),
        #         lty = c("dotted","solid","solid","dashed"),
        #         lwd = c(3,2,2,2), 
        #         name.pol = c("Obs Train station","Obs Test station", "Test results station", "Model"))
        #dev.off() #per eliminar la imatge un cop mostrada
        
        #Summary_stats <- openair::modStats(data_single, obs = "Obs_Test_station", mod = "Test_results_station")
        #print(Summary_stats)
        #summary(data_single)
        
        #analisi de la cross validation
        test_performace     <- postResample(pred = res_test, obs = data_test$obs)
        training_performace <- postResample(pred = res_training, obs = data_training$obs)
        
        #statistics
        #generem un df amb nomes les dades dels gaps i els resultats del gap-filling
        Results_Test <- data.frame(date = Obs_Test[, 'date'], Obs_Test = Obs_Test$Obs_Test, GBM = Test_results$Test_results)
        
        #calcul parametres estadistics
        Summary_stats_gbm <- openair::modStats(Results_Test, obs = "Obs_Test", mod = "GBM")
        
        #afegim (append) els parametres stats al df comu (FALTARIA FER PER DIFFERENTS GAP SIZES)
        Stats_gbm <- rbind(Stats_gbm, c(gap_size=gap_size, train_fraction = train_fraction, MB = Summary_stats_gbm$MB, r = Summary_stats_gbm$r, RMSE = Summary_stats_gbm$RMSE, seed = i_seed))
        
      }  
    }
  }
  
  
  
  
  
  if (s==13){
    #Eixample
    mean_gbm_eixample <- Stats_gbm %>%
      as.data.frame() %>%
      group_by(gap_size, train_fraction) %>%
      dplyr::summarise(r_mean= mean(r), r_std = sd(r), MB_mean= mean(MB), MB_std = sd(MB), RMSE_mean = mean(RMSE), RMSE_std = sd(RMSE))
    #saving
    saveRDS(object=mean_gbm_eixample, file=paste0(save_path,"mean_gbm_eixample-100.rds"))
    print("saved eixample")
  }else{
    #Senia 
    mean_gbm_senia <- Stats_gbm %>%
      as.data.frame() %>%
      group_by(gap_size, train_fraction) %>%
      dplyr::summarise(r_mean= mean(r), r_std = sd(r), MB_mean= mean(MB), MB_std = sd(MB), RMSE_mean = mean(RMSE), RMSE_std = sd(RMSE))
    saveRDS(object=mean_gbm_senia, file=paste0(save_path,"mean_gbm_senia-100.rds"))
    print("saved senia")
  }
}

end_time <- Sys.time()
time <- end_time - start_time
print(time)





