
############################################################
#                                                          #
#                          Libraries                       #
#                                                          #
############################################################
# 23-05-2023 PM2,5 de 2022
# Substitue all NA's in df_obs_sub by gbm predictions based on ok_total.

# Needs packages:
library(urbanR)
library(gbm)
library(lubridate)
library(plyr)
library(dplyr) #per treballar amb cbinds i esas cosas
library(caret)


#borrar
library(RNetCDF)
library(sp)
library(sf)
#library(mapview)
library(ncdf4)
library(rgeos) 
library(ggplot2)
#library(plyr) #OJO! la llibreria plyr ha de ser cridada abans que dplyr 
#library(dplyr) 
library(broom)
library(magrittr)
library(ggpubr)
library(raster)
library(rgdal)
#library(tmap)
library(gstat)
library(sp)
library(abind)
library(openair)
library(tidyr)
#library(caret) # an aggregator package for performing many machine learning models

############################################################
#                                                          #
#                    Initial data                          #
#                                                          #
############################################################
pollutant       <- "pm2p5"   #no2, o3, pm10
#years           <- c(2020, 2021, 2022)
years           <- 2019
save_result     <-"/esarchive/scratch/abarrant/results_regional_"   
file_format     <-"gene"   #ghost
time_resolution <-"daily"  #"hourly"
compute_in_UTM  <- "TRUE"
crs_latlon      <- "+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0"
crs_utm         <- "+proj=utm +zone=31 +ellps=intl +units=m +no_defs"

path_gbm    <- "/esarchive/scratch/abarrant/results_regional_2019/GBM_eval/"  ##change
############################################################
#                                                          #
#                 Load Local Stations                      #
#                                                          #
############################################################
#We will pick the latest data available
local_stations     <- readRDS(paste0(save_result,years[length(years)],"/results_gene_",pollutant, "/output/","local_stations_",file_format, "_",pollutant,".rds"))

############################################################
#                                                          #
#                  Load All observations                   #
#                                                          #
############################################################

df_obs_years <- NULL
for(read_year in years){
  df_obs_all        <- readRDS(file = paste0(save_result,read_year,"/results_gene_",pollutant, "/output/","df_obs_",file_format,"_",pollutant,".rds"))
  
  start_date  <- as.Date(paste(read_year, "-01-01", sep = ""))
  end_date    <- as.Date(paste(read_year, "-12-31", sep = ""))
  
  dates      <- seq(start_date, end_date, by = "day")                   
  df_obs_sub <- subset(df_obs_all, date>=start_date & date <= end_date) 
  
  
  # Comprovation we don't have duplications in obs
  if(any(duplicated(df_obs_sub$date))){
    print(paste0("Warning, there is duplicated observations on date :", df_obs_sub$date[duplicated(df_obs_sub$date)]))
    df_obs_sub <- distinct(df_obs_sub, date, .keep_all = TRUE)
    print("The df_obs_sub (alias=observations) have been modified to have the same length as the model")
  }
  
  if (length(df_obs_sub$date) != length(dates)){
    print("WARNING: There are missing hours/dates in observations. And the code is not developed for daily data yet!!")
  }
  
  df_obs_years <- rbind.fill(df_obs_years, df_obs_sub) #combine df by row, filling in missing columns
}

names_local_stations <- as.list(intersect(local_stations@data$ref_obs, colnames(df_obs_years)))

############################################################
#                                                          #
#            Load raw model interpolated to AQMS           #
#                                                          #
############################################################
df_all_model_raw_at_AQMS <- NULL
for(read_year in years){
  
  model_raw_at_AQMS_df     <- readRDS(paste0(save_result,read_year,"/results_gene_",pollutant, "/output/","model_raw_at_AQMS_",file_format, "_",pollutant,"_df.rds"))
  df_all_model_raw_at_AQMS <- rbind.fill(df_all_model_raw_at_AQMS, model_raw_at_AQMS_df)
}

############################################################
#                                                          #
#   AQMS interpolated to AQMS as covariate for GBM         #
#                                                          #
############################################################

# We pick up the observations configured in config file:
obs_filled       <- df_obs_years 
AQMS_obs         <- subset(df_obs_years, select = -date) #we remove the column dates from observations data frame

# To fill the gaps of observations in each station, so we need a loop to repeat the fill process in each station:
for (s in 1:length(names_local_stations)){
  #s <- 1
  #station                 <- names_local_stations[[s]]
  station <- "ES0567A"
  # Select station data (PREDICTORS)
  single_station_data <- data.frame(date            = df_all_model_raw_at_AQMS$date,
                                    obs             = AQMS_obs[,station],                     #valor obs a l'estacio (de vegades nan)
                                    model           = df_all_model_raw_at_AQMS[,station])              #predictor del valor interpolat a lestacio a partir de totes les altres 
  
  
  single_station_data                     <- na.omit(single_station_data)  #ometre predictors nan
  
  # Split train/test fractions 
  #train.fraction  <- 0.90
  train.fraction  <- 0.75
  smp_size        <- floor(train.fraction * nrow(single_station_data)) #te redondea el número i ya (129?)
  
  #seq_len et crea una seqüència de números nrow=47 seq_len de l'1 al 47
  #sample agafa un subconjunt aleatòri d'elements d'un vector, en aquest cas agafa el 90% (size)
  set.seed(16) #li he afegit, ns si te sentit que a cada estacio agafi el mateix train_id ---> Preguntar JAN
  train_ind     <- sample(seq_len(nrow(single_station_data)), size = smp_size) #42 valors aleatòris (90% de 48 h)
  data_training <- single_station_data[train_ind,]  %>% arrange(date)  #ordena les dates temporalment
  data_test     <- single_station_data[-train_ind,] %>% arrange(date)
  
  
  # Data frame amb dos variables date(fecha + hora) i Tests_results/Obs_Test/Obs_Train (NA) inicialitzacio df amb nan
  ## les dates son les del df single_station_data (ometent els nans intrinsecs de les estacions)
  Test_results_station  <-  data.frame(date=single_station_data$date , Test_results = NA) 
  Obs_Test_station      <-  data.frame(date=single_station_data$date , Obs_Test = NA)
  Obs_Train_station     <-  data.frame(date=single_station_data$date , Obs_Train = NA)
  
  
  ## creem un df amb nomes les dades de les obs train, i col·loquem les dades en el df creat amb totes les dates disponibles, deixant els forats nan
  Obs_Train <- data.frame(date = data_training[,'date'], Obs_Train=data_training$obs )
  Obs_Train_station[Obs_Train_station$date%in%Obs_Train$date ,"Obs_Train"] <- Obs_Train[,"Obs_Train"]
  
  ## cal generar les columnes buides?
  Obs_Persist <- data.frame(date = Obs_Train_station[,'date'], Obs_Persist= NA)
  Obs_Linterpol <- data.frame(date = Obs_Train_station[,'date'], Obs_Linterpol= NA)
  Obs_Train_station[, "Obs_Persist"] <- Obs_Persist[,"Obs_Persist"]
  Obs_Train_station[, "Obs_Linterpol"] <- Obs_Linterpol[,"Obs_Linterpol"]
  
  #funcio interpolacio lineal
  linear_interpol <- function(x) {
    y1 <- Obs_Train_station$Obs_Linterpol[index_above]
    y2 <- Obs_Train_station$Obs_Linterpol[index_behind + i - 1] #donat que agafem el menor index despres dels nan, i li hem de sumar i
    x1 <- index_above
    x2 <- index_behind + i - 1
    y1 + ((y2 - y1) * (x - x1)) / (x2 - x1)
  }
  
  ## corregir si el primer i ultim valor es nan!!!!
  for (i in 1:length(Obs_Train_station$date)) {
    if (!is.na(Obs_Train_station[i, "Obs_Train"])) {
      Obs_Train_station[i, "Obs_Persist"] <- Obs_Train_station[i, "Obs_Train"]     #si tenim el valor de training, simplement lagafem
      Obs_Train_station[i, "Obs_Linterpol"] <- Obs_Train_station[i, "Obs_Train"]
    } else {
      Obs_Train_station[i, "Obs_Persist"] <- Obs_Train_station[i - 1, "Obs_Persist"]  #si no tenim el valor, agafem l'ultim valor que tinguem (persistance)
      
      index_above <- max(which(!is.na(Obs_Train_station$Obs_Train[1:i])))                                #index primer valor abans dels nan
      index_behind <- min(which(!is.na(Obs_Train_station$Obs_Train[i:length(Obs_Train_station$date)])))  #index ultim valor despres nan
      
      #fem linterpolacio als valors nan
      Obs_Train_station[i, "Obs_Linterpol"] <- linear_interpol(i)
    }
  }

  ## afegim una columna amb els valors del test
  Obs_Test <- data.frame(date = data_test[,'date'], Obs_Test=data_test$obs )
  Obs_Train_station[Obs_Train_station$date%in%Obs_Test$date ,"Obs_Test"] <- Obs_Test[,"Obs_Test"]
  
  
  ## generem un df amb nomes els valors de test, afegim les col de persist i linterpol
  Results_Test <- data.frame(date = data_test[,'date'], Obs_Test=data_test$obs )
  Results_Test$Obs_Persist <- Obs_Train_station[Obs_Train_station$date%in%Obs_Test$date ,"Obs_Persist"]   
  Results_Test$Obs_Linterpol <- Obs_Train_station[Obs_Train_station$date%in%Obs_Test$date ,"Obs_Linterpol"]  
  
}

## calcul parametres estadistics
Summary_stats_persist    <- openair::modStats(Results_Test, obs = "Obs_Test", mod = "Obs_Persist")
Summary_stats_linterpol    <- openair::modStats(Results_Test, obs = "Obs_Test", mod = "Obs_Linterpol")
print(Summary_stats_persist)
print(Summary_stats_linterpol)


#saving
saveRDS(object=Results_Test, file=paste0(path_gbm,"benchmark_results.rds"))
benchmark_results <- readRDS(file=paste0(path_gbm,"benchmark_results.rds"))





